#include <Arduino.h>

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define AUTOTRIGGER_MAX_SAMPLES 10000

#define BUFF_LENGTH 1700

volatile uint8_t buffer[BUFF_LENGTH];
volatile uint8_t *pos = buffer;

volatile uint8_t trigged = 0;
volatile uint8_t dataReady = 0;
uint8_t triggerLevel = 128;
uint16_t preTriggerSamples = BUFF_LENGTH / 2;
uint16_t postTriggerSamples = BUFF_LENGTH -preTriggerSamples;
uint8_t triggerRising = 1;
uint8_t lastConnectionConfirmed = 0;
uint8_t lastTimeWasConnected = 0;
uint8_t terminalValuesNeedUpdate = 0;
uint8_t readyForNexMeasurement = 1;

uint8_t presetFsIndex = 3;
uint16_t presetTrigger_dv = 25;
uint8_t presetPreTrigger_x10percent = 5;
uint8_t prescalers[6] = {0b011, 0b011, 0b011, 0b011, 0b010, 0b010};
uint8_t counters[6] = {249, 124, 49, 24, 140 , 99};
uint16_t periods_us[6] = {1000, 500, 200, 100, 70 , 50};

enum Status
{
  paused,
  running,
  singleTrigger
};

enum Status status = running;

volatile uint16_t samplesRemaining;

void initADC()
{
  // Select Vref=AVcc
  //and set left adjust result
  ADMUX |= (1 << REFS0) | (1 << ADLAR);
  //set prescaller to 8
  //enable autotriggering
  //enable ADC interupt
  //and enable ADC
  ADCSRA = (1 << ADPS2) | (1 << ADATE) | (1 << ADIE) | (1 << ADEN);
  //set ADC trigger source - Timer0 compare match A
  cbi(ADCSRB, ADTS2);
  sbi(ADCSRB, ADTS1);
  sbi(ADCSRB, ADTS0);

  //select ADC channel with safety mask
  ADMUX = (ADMUX & 0xF0) | (0 & 0x0F);
}

void initTimer0(void)
{
  //Set Initial Timer value
  TCNT0 = 0;

  //Set CTC mode
  TCCR0A = (1 << WGM01);
}

void StartTimer0(void)
{
  //Place TOP timer value to Output compare register
  OCR0A = counters[presetFsIndex];
  //Set prescaller and start timer
  TCCR0B |= (prescalers[presetFsIndex]);
}

void StopTimer0(void)
{
  TCCR0B &= ~(0b111);
}

void StartADC(void)
{
  ADCSRA |= (1 << ADSC);
}

void buff_push_back(uint8_t value)
{
  if (!trigged && (samplesRemaining < AUTOTRIGGER_MAX_SAMPLES - preTriggerSamples))
  {
    if (triggerRising)
    {
      if (*pos < triggerLevel && value >= triggerLevel)
      {
        samplesRemaining = postTriggerSamples;
        trigged = 1;
      }
    }
    else
    {
      if (*pos > triggerLevel && value <= triggerLevel)
      {
        samplesRemaining = postTriggerSamples;
        trigged = 1;
      }
    }
  }
  if (++pos == buffer + BUFF_LENGTH)
    pos = buffer;
  *pos = value;
}

void beginMeasuring()
{
  samplesRemaining = AUTOTRIGGER_MAX_SAMPLES;
  pos = buffer;
  trigged = 0;
  preTriggerSamples = (BUFF_LENGTH * presetPreTrigger_x10percent) / 10;
  postTriggerSamples = BUFF_LENGTH - preTriggerSamples;
  triggerLevel = (256 * presetTrigger_dv) / 50;
  if ((presetFsIndex >= 3))                                                     // 10 to 20 kS
    ADCSRA = (1 << ADPS2) | (1 << ADATE) | (1 << ADIE) | (1 << ADEN);                //Prescaler 16
  else
    ADCSRA = (1 << ADPS2) | (1 << ADPS0) | (1 << ADATE) | (1 << ADIE) | (1 << ADEN); //Prescaler 32

  StartTimer0();
}

void drawTerminal()
{
  Serial.print(F("$$T\e[2JTrigger\e[2;1H\e[0m\e[2;4HV"));
  Serial.print(F("\e[2;6H\e[37;1m\e[48;5;214m\e[1m-+"));
  Serial.print(F("\e[0m\e[4;1HPretrigger\e[5;2H\e[1m"));
  Serial.print(F("\e[48;5;214m<\e[5;6H>\e[0m"));
  Serial.print(F("\e[0m\e[7;1HSampling\e[8;4H1 kHz\e[9;4H2 kHz"));
  Serial.print(F("\e[10;4H5 kHz\e[11;3H10 kHz\e[12;3H15 kHz"));
  Serial.print(F("\e[13;3H20 kHz"));
  Serial.print(F("\e[2;12H\e[0m\e[48;5;214m\e[1mRPS"));
}

void updateTerminalTrigLevel()
{
  Serial.print(F("$$T\e[2;1H\e[0m"));
  Serial.print(presetTrigger_dv / 10);
  Serial.print(F("."));
  Serial.print(presetTrigger_dv % 10);
  Serial.print(F("\e[0m"));
  Serial.print(F("$$Strigpos:"));
  Serial.print(presetTrigger_dv / 10);
  Serial.print(F("."));
  Serial.print(presetTrigger_dv % 10);
  Serial.print(F(";"));
}

void updateTerminalTrigEdge()
{
  if (triggerRising)
    Serial.print(F("$$T\e[2;9H\e[48;5;214m\e[37;1m/\e[0m"));
  else
    Serial.print(F("$$T\e[2;9H\e[48;5;214m\e[37;1m\\\e[0m"));
}

void updateTerminalPretrigger()
{
  if (presetPreTrigger_x10percent == 0)
    Serial.print(F("$$T\e[5;3H\e[0mOFF"));
  else
  {
    Serial.print(F("$$T\e[5;3H\e[0m"));
    Serial.print(presetPreTrigger_x10percent);
    Serial.print(F("0%"));
  }
}

void updateTerminalSample()
{
  Serial.print(F("$$T\e[8;1H\e[38;5;254m\e[48;5;254m0\e[9;1H1"));
  Serial.print(F("\e[10;1H2\e[11;1H3\e[12;1H4\e[13;1H5"));
  Serial.print(F("\e[38;5;214m\e[48;5;214m\e["));
  Serial.print(8 + presetFsIndex);
  Serial.print(F(";1H"));
  Serial.print((char)('0' + presetFsIndex));
}

void checkSerialRx()
{
  while (Serial.available() > 0)
  {
    char c;
    Serial.readBytes(&c, 1);

    switch (c)
    {
    case 'c':
      lastConnectionConfirmed = 1;
      break;
    case '+':
      if (presetTrigger_dv < 50)
      {
        presetTrigger_dv += 5;
        terminalValuesNeedUpdate = 1;
      }
      break;
    case '-':
      if (presetTrigger_dv > 5)
      {
        presetTrigger_dv -= 5;
        terminalValuesNeedUpdate = 1;
      }
      break;
    case '<':
      if (presetPreTrigger_x10percent > 0)
      {
        presetPreTrigger_x10percent -= 1;
        terminalValuesNeedUpdate = 1;
      }
      break;
    case '>':
      if (presetPreTrigger_x10percent < 9)
      {
        presetPreTrigger_x10percent += 1;
        terminalValuesNeedUpdate = 1;
      }
      break;
    case '/':
      triggerRising = 0;
      terminalValuesNeedUpdate = 1;
      break;
    case '\\':
      triggerRising = 1;
      terminalValuesNeedUpdate = 1;
      break;
    case 'R':
      status = running;
      terminalValuesNeedUpdate = 1;
      break;
    case 'P':
      status = paused;
      terminalValuesNeedUpdate = 1;
      break;
    case 'S':
      status = singleTrigger;
      terminalValuesNeedUpdate = 1;
      break;

    default:
      if (c >= '0' && c <= '5')
      {
        presetFsIndex = c - '0';
        terminalValuesNeedUpdate = 1;
      }
      break;
    }
  }
}

int main()
{
  Serial.begin(250000);

  initADC();

  initTimer0();

  sei(); //Interrupt global enable

  StartADC();

  beginMeasuring();

  while (1)
  {
    checkSerialRx();

    if (dataReady)
    {
      if (status == singleTrigger)
      {
        status = paused;
        terminalValuesNeedUpdate = 1;
      }

      if (lastConnectionConfirmed)
      {
        if (!lastTimeWasConnected)
        {
          Serial.print(F("$$IConnection detected"));
          drawTerminal();
          terminalValuesNeedUpdate = 1;
        }

        Serial.print(F("$$C1,"));
        Serial.print(periods_us[presetFsIndex]);
        Serial.print(F("e-6,"));
        Serial.print(BUFF_LENGTH);
        Serial.print(F(",8,0,5,"));
        Serial.print(preTriggerSamples);
        Serial.print(F(";u1"));

        uint16_t firstSampleIndex = (++pos) - buffer;
        if (firstSampleIndex == BUFF_LENGTH)
          firstSampleIndex = 0;

        Serial.write((uint8_t *)&buffer[firstSampleIndex], (BUFF_LENGTH - firstSampleIndex));
        Serial.write((uint8_t *)&buffer[0], (firstSampleIndex));
        Serial.print(F(";"));

        if (trigged)
          Serial.print(F("$$T\e[0m\e[1;9H\e[48;5;34mNORMAL"));
        else
          Serial.print(F("$$T\e[0m\e[1;9H\e[41;1m AUTO "));
      }

      dataReady = 0;
      readyForNexMeasurement = 1;

      lastTimeWasConnected = lastConnectionConfirmed;
      lastConnectionConfirmed=0;
      Serial.print(F("$$Ec"));
    }

    if (status == singleTrigger)
      Serial.print(F("$$T\e[0m\e[1;9H\e[41;1m WAIT "));

    if (status == paused)
      Serial.print(F("$$T\e[0m\e[1;9H\e[41;1mPAUSED"));

    if (terminalValuesNeedUpdate)
    {
      terminalValuesNeedUpdate = 0;

      updateTerminalTrigEdge();
      updateTerminalTrigLevel();
      updateTerminalPretrigger();
      updateTerminalSample();
    }

    if (readyForNexMeasurement)
    {
      if (status != paused)
      {
        readyForNexMeasurement = 0;
        beginMeasuring();
      }
    }
  }
}

ISR(ADC_vect)
{
  uint8_t high = ADCH;
  buff_push_back(high);

  samplesRemaining--;

  if (samplesRemaining == 0)
  {
    if (status == singleTrigger)
    {
      if (trigged)
        goto stop;
      else
        samplesRemaining = AUTOTRIGGER_MAX_SAMPLES;
    }
    else
    {
    stop:
      StopTimer0();
      dataReady = 1;
    }
  }

  //clear timer compare match flag
  TIFR0 = (1 << OCF0A);
}